@@ -2,6 +2,7 @@
 #define TSRM_CONFIG_W32_H
 
 #include <../main/config.w32.h>
+#include "Zend/zend_config.w32.h"
 
 #define HAVE_UTIME 1
 #define HAVE_ALLOCA 1
@@ -11,12 +12,13 @@
 #include <stdlib.h>
 #include <crtdbg.h>
 
-#undef inline
-#ifdef ZEND_WIN32_FORCE_INLINE
-# define inline __forceinline
-#else
-# define inline
 #endif
 
-
-#endif
+/*
+ * Local variables:
+ * tab-width: 4
+ * c-basic-offset: 4
+ * End:
+ * vim600: sw=4 ts=4 fdm=marker
+ * vim<600: sw=4 ts=4
+ */
