@@ -4,6 +4,7 @@
 #include <stdlib.h>
 #include "gd.h"
 #include "gdhelpers.h"
+#include "gd_errors.h"
 
 #include "php.h"
 
@@ -91,33 +92,82 @@ static const unsigned char gd_toascii[256] =
 extern int gdCosT[];
 extern int gdSinT[];
 
+/**
+ * Group: Error Handling
+ */
+
+void gd_stderr_error(int priority, const char *format, va_list args)
+{
+	switch (priority) {
+	case GD_ERROR:
+		fputs("GD Error: ", stderr);
+		break;
+	case GD_WARNING:
+		fputs("GD Warning: ", stderr);
+		break;
+	case GD_NOTICE:
+		fputs("GD Notice: ", stderr);
+		break;
+#ifndef PHP_WIN32
+	case GD_INFO:
+		fputs("GD Info: ", stderr);
+		break;
+	case GD_DEBUG:
+		fputs("GD Debug: ", stderr);
+		break;
+#endif
+	}
+	vfprintf(stderr, format, args);
+	fflush(stderr);
+}
+
+static gdErrorMethod gd_error_method = gd_stderr_error;
+
+static void _gd_error_ex(int priority, const char *format, va_list args)
+{
+	if (gd_error_method) {
+		gd_error_method(priority, format, args);
+	}
+}
+
+void gd_error(const char *format, ...)
+{
+	va_list args;
+
+	va_start(args, format);
+	_gd_error_ex(GD_WARNING, format, args);
+	va_end(args);
+}
+void gd_error_ex(int priority, const char *format, ...)
+{
+	va_list args;
+
+	va_start(args, format);
+	_gd_error_ex(priority, format, args);
+	va_end(args);
+}
+
+/*
+	Function: gdSetErrorMethod
+*/
+void gdSetErrorMethod(gdErrorMethod error_method)
+{
+	gd_error_method = error_method;
+}
+
+/*
+	Function: gdClearErrorMethod
+*/
+void gdClearErrorMethod(void)
+{
+	gd_error_method = gd_stderr_error;
+}
+
 static void gdImageBrushApply(gdImagePtr im, int x, int y);
 static void gdImageTileApply(gdImagePtr im, int x, int y);
-static void gdImageAntiAliasedApply(gdImagePtr im, int x, int y);
-static int gdLayerOverlay(int dst, int src);
 static int gdAlphaOverlayColor(int src, int dst, int max);
 int gdImageGetTrueColorPixel(gdImagePtr im, int x, int y);
 
-void php_gd_error_ex(int type, const char *format, ...)
-{
-	va_list args;
-
-
-	va_start(args, format);
-	php_verror(NULL, "", type, format, args);
-	va_end(args);
-}
-
-void php_gd_error(const char *format, ...)
-{
-	va_list args;
-
-
-	va_start(args, format);
-	php_verror(NULL, "", E_WARNING, format, args);
-	va_end(args);
-}
-
 gdImagePtr gdImageCreate (int sx, int sy)
 {
 	int i;
@@ -126,12 +176,10 @@ gdImagePtr gdImageCreate (int sx, int sy)
 	if (overflow2(sx, sy)) {
 		return NULL;
 	}
-
 	if (overflow2(sizeof(unsigned char *), sy)) {
 		return NULL;
 	}
-
-	if (overflow2(sizeof(unsigned char *), sx)) {
+	if (overflow2(sizeof(unsigned char), sx)) {
 		return NULL;
 	}
 
@@ -139,7 +187,6 @@ gdImagePtr gdImageCreate (int sx, int sy)
 
 	/* Row-major ever since gd 1.3 */
 	im->pixels = (unsigned char **) gdMalloc(sizeof(unsigned char *) * sy);
-	im->AA_opacity = (unsigned char **) gdMalloc(sizeof(unsigned char *) * sy);
 	im->polyInts = 0;
 	im->polyAllocated = 0;
 	im->brush = 0;
@@ -148,7 +195,6 @@ gdImagePtr gdImageCreate (int sx, int sy)
 	for (i = 0; i < sy; i++) {
 		/* Row-major ever since gd 1.3 */
 		im->pixels[i] = (unsigned char *) gdCalloc(sx, sizeof(unsigned char));
-		im->AA_opacity[i] = (unsigned char *) gdCalloc(sx, sizeof(unsigned char));
 	}
 	im->sx = sx;
 	im->sy = sy;
@@ -157,7 +203,6 @@ gdImagePtr gdImageCreate (int sx, int sy)
 	im->interlace = 0;
 	im->thick = 1;
 	im->AA = 0;
-	im->AA_polygon = 0;
 	for (i = 0; i < gdMaxColors; i++) {
 		im->open[i] = 1;
 		im->red[i] = 0;
@@ -170,6 +215,8 @@ gdImagePtr gdImageCreate (int sx, int sy)
 	im->cy1 = 0;
 	im->cx2 = im->sx - 1;
 	im->cy2 = im->sy - 1;
+	im->res_x = GD_RESOLUTION;
+	im->res_y = GD_RESOLUTION;
 	im->interpolation = NULL;
 	im->interpolation_id = GD_BILINEAR_FIXED;
 	return im;
@@ -183,11 +230,9 @@ gdImagePtr gdImageCreateTrueColor (int sx, int sy)
 	if (overflow2(sx, sy)) {
 		return NULL;
 	}
-
-	if (overflow2(sizeof(unsigned char *), sy)) {
+	if (overflow2(sizeof(int *), sy)) {
 		return NULL;
 	}
-
 	if (overflow2(sizeof(int), sx)) {
 		return NULL;
 	}
@@ -195,7 +240,6 @@ gdImagePtr gdImageCreateTrueColor (int sx, int sy)
 	im = (gdImage *) gdMalloc(sizeof(gdImage));
 	memset(im, 0, sizeof(gdImage));
 	im->tpixels = (int **) gdMalloc(sizeof(int *) * sy);
-	im->AA_opacity = (unsigned char **) gdMalloc(sizeof(unsigned char *) * sy);
 	im->polyInts = 0;
 	im->polyAllocated = 0;
 	im->brush = 0;
@@ -203,7 +247,6 @@ gdImagePtr gdImageCreateTrueColor (int sx, int sy)
 	im->style = 0;
 	for (i = 0; i < sy; i++) {
 		im->tpixels[i] = (int *) gdCalloc(sx, sizeof(int));
-		im->AA_opacity[i] = (unsigned char *) gdCalloc(sx, sizeof(unsigned char));
 	}
 	im->sx = sx;
 	im->sy = sy;
@@ -220,11 +263,12 @@ gdImagePtr gdImageCreateTrueColor (int sx, int sy)
 	im->alphaBlendingFlag = 1;
 	im->thick = 1;
 	im->AA = 0;
-	im->AA_polygon = 0;
 	im->cx1 = 0;
 	im->cy1 = 0;
 	im->cx2 = im->sx - 1;
 	im->cy2 = im->sy - 1;
+	im->res_x = GD_RESOLUTION;
+	im->res_y = GD_RESOLUTION;
 	im->interpolation = NULL;
 	im->interpolation_id = GD_BILINEAR_FIXED;
 	return im;
@@ -245,12 +289,6 @@ void gdImageDestroy (gdImagePtr im)
 		}
 		gdFree(im->tpixels);
 	}
-	if (im->AA_opacity) {
-		for (i = 0; i < im->sy; i++) {
-			gdFree(im->AA_opacity[i]);
-		}
-		gdFree(im->AA_opacity);
-	}
 	if (im->polyInts) {
 		gdFree(im->polyInts);
 	}
@@ -597,15 +635,18 @@ void gdImageColorDeallocate (gdImagePtr im, int color)
 
 void gdImageColorTransparent (gdImagePtr im, int color)
 {
+	if (color < 0) {
+		return;
+	}
 	if (!im->trueColor) {
+		if((color >= im->colorsTotal)) {
+			return;
+		}
+		/* Make the old transparent color opaque again */
 		if (im->transparent != -1) {
 			im->alpha[im->transparent] = gdAlphaOpaque;
 		}
-		if (color > -1 && color < im->colorsTotal && color < gdMaxColors) {
-			im->alpha[color] = gdAlphaTransparent;
-		} else {
-			return;
-		}
+		im->alpha[color] = gdAlphaTransparent;
 	}
 	im->transparent = color;
 }
@@ -753,7 +794,9 @@ void gdImageSetPixel (gdImagePtr im, int x, int y, int color)
 			gdImageTileApply(im, x, y);
 			break;
 		case gdAntiAliased:
-			gdImageAntiAliasedApply(im, x, y);
+			/* This shouldn't happen (2.0.26) because we just call
+			  gdImageAALine now, but do something sane. */
+			gdImageSetPixel(im, x, y, im->AA_color);
 			break;
 		default:
 			if (gdImageBoundsSafe(im, x, y)) {
@@ -764,14 +807,15 @@ void gdImageSetPixel (gdImagePtr im, int x, int y, int color)
 							im->tpixels[y][x] = color;
 							break;
 						case gdEffectAlphaBlend:
-							im->tpixels[y][x] = gdAlphaBlend(im->tpixels[y][x], color);
-							break;
 						case gdEffectNormal:
 							im->tpixels[y][x] = gdAlphaBlend(im->tpixels[y][x], color);
 							break;
 						case gdEffectOverlay :
 							im->tpixels[y][x] = gdLayerOverlay(im->tpixels[y][x], color);
 							break;
+						case gdEffectMultiply :
+							im->tpixels[y][x] = gdLayerMultiply(im->tpixels[y][x], color);
+							break;
 					}
 				} else {
 					im->pixels[y][x] = color;
@@ -934,68 +978,6 @@ static int gdImageTileGet (gdImagePtr im, int x, int y)
 }
 
 
-static void gdImageAntiAliasedApply (gdImagePtr im, int px, int py)
-{
-	float p_dist, p_alpha;
-	unsigned char opacity;
-
-	/*
-	 * Find the perpendicular distance from point C (px, py) to the line
-	 * segment AB that is being drawn.  (Adapted from an algorithm from the
-	 * comp.graphics.algorithms FAQ.)
-	 */
-
-	int LAC_2, LBC_2;
-
-	int Ax_Cx = im->AAL_x1 - px;
-	int Ay_Cy = im->AAL_y1 - py;
-
-	int Bx_Cx = im->AAL_x2 - px;
-	int By_Cy = im->AAL_y2 - py;
-
-	/* 2.0.13: bounds check! AA_opacity is just as capable of
-	 * overflowing as the main pixel array. Arne Jorgensen.
-	 * 2.0.14: typo fixed. 2.0.15: moved down below declarations
-	 * to satisfy non-C++ compilers.
-	 */
-	if (!gdImageBoundsSafe(im, px, py)) {
-		return;
-	}
-
-	/* Get the squares of the lengths of the segemnts AC and BC. */
-	LAC_2 = (Ax_Cx * Ax_Cx) + (Ay_Cy * Ay_Cy);
-	LBC_2 = (Bx_Cx * Bx_Cx) + (By_Cy * By_Cy);
-
-	if (((im->AAL_LAB_2 + LAC_2) >= LBC_2) && ((im->AAL_LAB_2 + LBC_2) >= LAC_2)) {
-		/* The two angles are acute.  The point lies inside the portion of the
-		 * plane spanned by the line segment.
-		 */
-		p_dist = fabs ((float) ((Ay_Cy * im->AAL_Bx_Ax) - (Ax_Cx * im->AAL_By_Ay)) / im->AAL_LAB);
-	} else {
-		/* The point is past an end of the line segment.  It's length from the
-		 * segment is the shorter of the lengths from the endpoints, but call
-		 * the distance -1, so as not to compute the alpha nor draw the pixel.
-		 */
-		p_dist = -1;
-	}
-
-	if ((p_dist >= 0) && (p_dist <= (float) (im->thick))) {
-		p_alpha = pow (1.0 - (p_dist / 1.5), 2);
-
-		if (p_alpha > 0) {
-			if (p_alpha >= 1) {
-				opacity = 255;
-			} else {
-				opacity = (unsigned char) (p_alpha * 255.0);
-			}
-			if (!im->AA_polygon || (im->AA_opacity[py][px] < opacity)) {
-				im->AA_opacity[py][px] = opacity;
-			}
-		}
-	}
-}
-
-
 int gdImageGetPixel (gdImagePtr im, int x, int y)
 {
 	if (gdImageBoundsSafe(im, x, y)) {
@@ -1011,53 +993,16 @@ int gdImageGetPixel (gdImagePtr im, int x, int y)
 
 void gdImageAABlend (gdImagePtr im)
 {
-	float p_alpha, old_alpha;
-	int color = im->AA_color, color_red, color_green, color_blue;
-	int old_color, old_red, old_green, old_blue;
-	int p_color, p_red, p_green, p_blue;
-	int px, py;
-
-	color_red = gdImageRed(im, color);
-	color_green = gdImageGreen(im, color);
-	color_blue = gdImageBlue(im, color);
-
-	/* Impose the anti-aliased drawing on the image. */
-	for (py = 0; py < im->sy; py++) {
-		for (px = 0; px < im->sx; px++) {
-			if (im->AA_opacity[py][px] != 0) {
-				old_color = gdImageGetPixel(im, px, py);
-
-				if ((old_color != color) && ((old_color != im->AA_dont_blend) || (im->AA_opacity[py][px] == 255))) {
-					/* Only blend with different colors that aren't the dont_blend color. */
-					p_alpha = (float) (im->AA_opacity[py][px]) / 255.0;
-					old_alpha = 1.0 - p_alpha;
-
-					if (p_alpha >= 1.0) {
-						p_color = color;
-					} else {
-						old_red = gdImageRed(im, old_color);
-						old_green = gdImageGreen(im, old_color);
-						old_blue = gdImageBlue(im, old_color);
-
-						p_red = (int) (((float) color_red * p_alpha) + ((float) old_red * old_alpha));
-						p_green = (int) (((float) color_green * p_alpha) + ((float) old_green * old_alpha));
-						p_blue = (int) (((float) color_blue * p_alpha) + ((float) old_blue * old_alpha));
-						p_color = gdImageColorResolve(im, p_red, p_green, p_blue);
-					}
-					gdImageSetPixel(im, px, py, p_color);
-				}
-			}
-		}
-		/* Clear the AA_opacity array behind us. */
-		memset(im->AA_opacity[py], 0, im->sx);
-	}
+	(void)im;
 }
 
+static void _gdImageFilledHRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color);
+
 static void gdImageHLine(gdImagePtr im, int y, int x1, int x2, int col)
 {
 	if (im->thick > 1) {
 		int thickhalf = im->thick >> 1;
-		gdImageFilledRectangle(im, x1, y - thickhalf, x2, y + im->thick - thickhalf - 1, col);
+		_gdImageFilledHRectangle(im, x1, y - thickhalf, x2, y + im->thick - thickhalf - 1, col);
 	} else {
 		if (x2 < x1) {
 			int t = x2;
@@ -1110,7 +1055,7 @@ void gdImageLine (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
 	}
 
 	/* 2.0.10: Nick Atty: clip to edges of drawing rectangle, return if no points need to be drawn */
-	if (!clip_1d(&x1,&y1,&x2,&y2,gdImageSX(im)) || !clip_1d(&y1,&x1,&y2,&x2,gdImageSY(im))) {
+	if (!clip_1d(&x1,&y1,&x2,&y2,gdImageSX(im)-1) || !clip_1d(&y1,&x1,&y2,&x2,gdImageSY(im)-1)) {
 		return;
 	}
 
@@ -1291,58 +1236,19 @@ inline static void gdImageSetAAPixelColor(gdImagePtr im, int x, int y, int color
 void gdImageAALine (gdImagePtr im, int x1, int y1, int x2, int y2, int col)
 {
 	/* keep them as 32bits */
-	long x, y, inc;
+	long x, y, inc, frac;
 	long dx, dy,tmp;
 
-	if (y1 < 0 && y2 < 0) {
+	if (!im->trueColor) {
+		/* TBB: don't crash when the image is of the wrong type */
+		gdImageLine(im, x1, y1, x2, y2, col);
 		return;
 	}
-	if (y1 < 0) {
-		x1 += (y1 * (x1 - x2)) / (y2 - y1);
-		y1 = 0;
-	}
-	if (y2 < 0) {
-		x2 += (y2 * (x1 - x2)) / (y2 - y1);
-		y2 = 0;
-	}
 
-	/* bottom edge */
-	if (y1 >= im->sy && y2 >= im->sy) {
+	/* 2.0.10: Nick Atty: clip to edges of drawing rectangle, return if no points need to be drawn */
+	if (!clip_1d(&x1,&y1,&x2,&y2,gdImageSX(im)-1) || !clip_1d(&y1,&x1,&y2,&x2,gdImageSY(im)-1)) {
 		return;
 	}
-	if (y1 >= im->sy) {
-		x1 -= ((im->sy - y1) * (x1 - x2)) / (y2 - y1);
-		y1 = im->sy - 1;
-	}
-	if (y2 >= im->sy) {
-		x2 -= ((im->sy - y2) * (x1 - x2)) / (y2 - y1);
-		y2 = im->sy - 1;
-	}
-
-	/* left edge */
-	if (x1 < 0 && x2 < 0) {
-		return;
-	}
-	if (x1 < 0) {
-		y1 += (x1 * (y1 - y2)) / (x2 - x1);
-		x1 = 0;
-	}
-	if (x2 < 0) {
-		y2 += (x2 * (y1 - y2)) / (x2 - x1);
-		x2 = 0;
-	}
-	/* right edge */
-	if (x1 >= im->sx && x2 >= im->sx) {
-		return;
-	}
-	if (x1 >= im->sx) {
-		y1 -= ((im->sx - x1) * (y1 - y2)) / (x2 - x1);
-		x1 = im->sx - 1;
-	}
-	if (x2 >= im->sx) {
-		y2 -= ((im->sx - x2) * (y1 - y2)) / (x2 - x1);
-		x2 = im->sx - 1;
-	}
 
 	dx = x2 - x1;
 	dy = y2 - y1;
@@ -1361,16 +1267,22 @@ void gdImageAALine (gdImagePtr im, int x1, int y1, int x2, int y2, int col)
 			dx = x2 - x1;
 			dy = y2 - y1;
 		}
-		x = x1 << 16;
-		y = y1 << 16;
+		y = y1;
 		inc = (dy * 65536) / dx;
-		while ((x >> 16) <= x2) {
-			gdImageSetAAPixelColor(im, x >> 16, y >> 16, col, (y >> 8) & 0xFF);
-			if ((y >> 16) + 1 < im->sy) {
-				gdImageSetAAPixelColor(im, x >> 16, (y >> 16) + 1,col, (~y >> 8) & 0xFF);
+		frac = 0;
+		for (x = x1; x <= x2; x++) {
+			gdImageSetAAPixelColor(im, x, y, col, (frac >> 8) & 0xFF);
+			if (y + 1 < im->sy) {
+				gdImageSetAAPixelColor(im, x, y + 1, col, (~frac >> 8) & 0xFF);
+			}
+			frac += inc;
+			if (frac >= 65536) {
+				frac -= 65536;
+				y++;
+			} else if (frac < 0) {
+				frac += 65536;
+				y--;
 			}
-			x += (1 << 16);
-			y += inc;
 		}
 	} else {
 		if (dy < 0) {
@@ -1383,16 +1295,22 @@ void gdImageAALine (gdImagePtr im, int x1, int y1, int x2, int y2, int col)
 			dx = x2 - x1;
 			dy = y2 - y1;
 		}
-		x = x1 << 16;
-		y = y1 << 16;
+		x = x1;
 		inc = (dx * 65536) / dy;
-		while ((y>>16) <= y2) {
-			gdImageSetAAPixelColor(im, x >> 16, y >> 16, col, (x >> 8) & 0xFF);
-			if ((x >> 16) + 1 < im->sx) {
-				gdImageSetAAPixelColor(im, (x >> 16) + 1, (y >> 16),col, (~x >> 8) & 0xFF);
+		frac = 0;
+		for (y = y1; y <= y2; y++) {
+			gdImageSetAAPixelColor(im, x, y, col, (frac >> 8) & 0xFF);
+			if (x + 1 < im->sx) {
+				gdImageSetAAPixelColor(im, x + 1, y, col, (~frac >> 8) & 0xFF);
+			}
+			frac += inc;
+			if (frac >= 65536) {
+				frac -= 65536;
+				x++;
+			} else if (frac < 0) {
+				frac += 65536;
+				x--;
 			}
-			x += inc;
-			y += (1<<16);
 		}
 	}
 }
@@ -1663,23 +1581,17 @@ long lsqrt (long n)
 /* s and e are integers modulo 360 (degrees), with 0 degrees
    being the rightmost extreme and degrees changing clockwise.
    cx and cy are the center in pixels; w and h are the horizontal
-   and vertical diameter in pixels. Nice interface, but slow.
-   See gd_arc_f_buggy.c for a better version that doesn't
-   seem to be bug-free yet. */
+   and vertical diameter in pixels. */
 
 void gdImageArc (gdImagePtr im, int cx, int cy, int w, int h, int s, int e, int color)
 {
-	if ((s % 360) == (e % 360)) {
-		gdImageEllipse(im, cx, cy, w, h, color);
-	} else {
-		gdImageFilledArc(im, cx, cy, w, h, s, e, color, gdNoFill);
-	}
+	gdImageFilledArc(im, cx, cy, w, h, s, e, color, gdNoFill);
 }
 
 void gdImageFilledArc (gdImagePtr im, int cx, int cy, int w, int h, int s, int e, int color, int style)
 {
-	gdPoint pts[3];
-	int i;
+	gdPoint pts[363];
+	int i, pti;
 	int lx = 0, ly = 0;
 	int fx = 0, fy = 0;
 
@@ -1707,7 +1619,7 @@ void gdImageFilledArc (gdImagePtr im, int cx, int cy, int w, int h, int s, int e
 		}
 	}
 
-	for (i = s; i <= e; i++) {
+	for (i = s, pti = 1; i <= e; i++, pti++) {
 		int x, y;
 		x = ((long) gdCosT[i % 360] * (long) w / (2 * 1024)) + cx;
 		y = ((long) gdSinT[i % 360] * (long) h / (2 * 1024)) + cy;
@@ -1716,19 +1628,28 @@ void gdImageFilledArc (gdImagePtr im, int cx, int cy, int w, int h, int s, int e
 				if (style & gdNoFill) {
 					gdImageLine(im, lx, ly, x, y, color);
 				} else {
-					/* This is expensive! */
-					pts[0].x = lx;
-					pts[0].y = ly;
-					pts[1].x = x;
-					pts[1].y = y;
-					pts[2].x = cx;
-					pts[2].y = cy;
-					gdImageFilledPolygon(im, pts, 3, color);
-				}
+					if (y == ly) {
+						pti--; /* don't add this point */
+						if (((i > 270 || i < 90) && x > lx) || ((i >  90 && i < 270) && x < lx)) {
+							/* replace the old x coord, if increasing on the
+							   right side or decreasing on the left side */
+							pts[pti].x = x;
+						}
+					} else {
+						pts[pti].x = x;
+						pts[pti].y = y;
+					}
+  				}
 			}
 		} else {
 			fx = x;
 			fy = y;
+			if (!(style & (gdChord | gdNoFill))) {
+				pts[0].x = cx;
+				pts[0].y = cy;
+				pts[pti].x = x;
+				pts[pti].y = y;
+			}
 		}
 		lx = x;
 		ly = y;
@@ -1755,10 +1676,108 @@ void gdImageFilledArc (gdImagePtr im, int cx, int cy, int w, int h, int s, int e
 				gdImageLine(im, cx, cy, lx, ly, color);
 				gdImageLine(im, cx, cy, fx, fy, color);
 			}
+		} else {
+			pts[pti].x = cx;
+			pts[pti].y = cy;
+			gdImageFilledPolygon(im, pts, pti+1, color);
 		}
 	}
 }
 
+/**
+ * Integer Ellipse functions (gdImageEllipse and gdImageFilledEllipse)
+ * Function added by Pierre-Alain Joye 02/08/2003 (paj@pearfr.org)
+ * See the ellipse function simplification for the equation
+ * as well as the midpoint algorithm.
+ */
+
+void gdImageEllipse(gdImagePtr im, int mx, int my, int w, int h, int c)
+{
+	int x=0,mx1=0,mx2=0,my1=0,my2=0;
+	long aq,bq,dx,dy,r,rx,ry,a,b;
+
+	a=w>>1;
+	b=h>>1;
+	gdImageSetPixel(im,mx+a, my, c);
+	gdImageSetPixel(im,mx-a, my, c);
+	mx1 = mx-a;my1 = my;
+	mx2 = mx+a;my2 = my;
+
+	aq = a * a;
+	bq = b * b;
+	dx = aq << 1;
+	dy = bq << 1;
+	r  = a * bq;
+	rx = r << 1;
+	ry = 0;
+	x = a;
+	while (x > 0){
+		if (r > 0) {
+			my1++;my2--;
+			ry +=dx;
+			r  -=ry;
+		}
+		if (r <= 0){
+			x--;
+			mx1++;mx2--;
+			rx -=dy;
+			r  +=rx;
+		}
+		gdImageSetPixel(im,mx1, my1, c);
+		gdImageSetPixel(im,mx1, my2, c);
+		gdImageSetPixel(im,mx2, my1, c);
+		gdImageSetPixel(im,mx2, my2, c);
+	}
+}
+
+void gdImageFilledEllipse (gdImagePtr im, int mx, int my, int w, int h, int c)
+{
+	int x=0,mx1=0,mx2=0,my1=0,my2=0;
+	long aq,bq,dx,dy,r,rx,ry,a,b;
+	int i;
+	int old_y2;
+
+	a=w>>1;
+	b=h>>1;
+
+	for (x = mx-a; x <= mx+a; x++) {
+		gdImageSetPixel(im, x, my, c);
+	}
+
+	mx1 = mx-a;my1 = my;
+	mx2 = mx+a;my2 = my;
+
+	aq = a * a;
+	bq = b * b;
+	dx = aq << 1;
+	dy = bq << 1;
+	r  = a * bq;
+	rx = r << 1;
+	ry = 0;
+	x = a;
+	old_y2=-2;
+	while (x > 0){
+		if (r > 0) {
+			my1++;my2--;
+			ry +=dx;
+			r  -=ry;
+		}
+		if (r <= 0){
+			x--;
+			mx1++;mx2--;
+			rx -=dy;
+			r  +=rx;
+		}
+		if(old_y2!=my2){
+			for(i=mx1;i<=mx2;i++){
+				gdImageSetPixel(im,i,my1,c);
+				gdImageSetPixel(im,i,my2,c);
+			}
+		}
+		old_y2 = my2;
+	}
+}
+
 void gdImageFillToBorder (gdImagePtr im, int x, int y, int border, int color)
 {
 	int lastBorder;
@@ -1766,7 +1785,7 @@ void gdImageFillToBorder (gdImagePtr im, int x, int y, int border, int color)
 	int leftLimit = -1, rightLimit;
 	int i, restoreAlphaBlending = 0;
 
-	if (border < 0) {
+	if (border < 0 || color < 0) {
 		/* Refuse to fill to a non-solid border */
 		return;
 	}
@@ -2043,7 +2062,6 @@ skip:		for(x++; x<=x2 && (pts[y][x] || gdImageGetPixel(im,x, y)!=oc); x++);
 
 void gdImageRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
 {
-	int x1h = x1, x1v = x1, y1h = y1, y1v = y1, x2h = x2, x2v = x2, y2h = y2, y2v = y2;
 	int thick = im->thick;
 	int t;
 
@@ -2064,7 +2082,6 @@ void gdImageRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
 		x2 = t;
 	}
 
-	x1h = x1; x1v = x1; y1h = y1; y1v = y1; x2h = x2; x2v = x2; y2h = y2; y2v = y2;
 	if (thick > 1) {
 		int cx, cy, x1ul, y1ul, x2lr, y2lr;
 		int half = thick >> 1;
@@ -2112,20 +2129,61 @@ void gdImageRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
 		if (x1 == x2 || y1 == y2) {
 			gdImageLine(im, x1, y1, x2, y2, color);
 		} else {
-			y1v = y1h + 1;
-			y2v = y2h - 1;
-			gdImageLine(im, x1h, y1h, x2h, y1h, color);
-			gdImageLine(im, x1h, y2h, x2h, y2h, color);
-			gdImageLine(im, x1v, y1v, x1v, y2v, color);
-			gdImageLine(im, x2v, y1v, x2v, y2v, color);
+			gdImageLine(im, x1, y1, x2, y1, color);
+			gdImageLine(im, x1, y2, x2, y2, color);
+			gdImageLine(im, x1, y1 + 1, x1, y2 - 1, color);
+			gdImageLine(im, x2, y1 + 1, x2, y2 - 1, color);
 		}
 	}
 }
 
-void gdImageFilledRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
+static void _gdImageFilledHRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
 {
 	int x, y;
 
+	if (x1 == x2 && y1 == y2) {
+		gdImageSetPixel(im, x1, y1, color);
+		return;
+	}
+
+	if (x1 > x2) {
+		x = x1;
+		x1 = x2;
+		x2 = x;
+	}
+
+	if (y1 > y2) {
+		y = y1;
+		y1 = y2;
+		y2 = y;
+	}
+
+	if (x1 < 0) {
+		x1 = 0;
+	}
+
+	if (x2 >= gdImageSX(im)) {
+		x2 = gdImageSX(im) - 1;
+	}
+
+	if (y1 < 0) {
+		y1 = 0;
+	}
+
+	if (y2 >= gdImageSY(im)) {
+		y2 = gdImageSY(im) - 1;
+	}
+
+	for (x = x1; (x <= x2); x++) {
+		for (y = y1; (y <= y2); y++) {
+			gdImageSetPixel (im, x, y, color);
+		}
+	}
+}
+
+static void _gdImageFilledVRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
+{
+	int x, y;
 
 	if (x1 == x2 && y1 == y2) {
 		gdImageSetPixel(im, x1, y1, color);
@@ -2167,6 +2225,11 @@ void gdImageFilledRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int
 	}
 }
 
+void gdImageFilledRectangle (gdImagePtr im, int x1, int y1, int x2, int y2, int color)
+{
+	_gdImageFilledVRectangle(im, x1, y1, x2, y2, color);
+}
+
 void gdImageCopy (gdImagePtr dst, gdImagePtr src, int dstX, int dstY, int srcX, int srcY, int w, int h)
 {
 	int c;
@@ -2185,7 +2248,9 @@ void gdImageCopy (gdImagePtr dst, gdImagePtr src, int dstX, int dstY, int srcX,
 			for (y = 0; (y < h); y++) {
 				for (x = 0; (x < w); x++) {
 					int c = gdImageGetTrueColorPixel (src, srcX + x, srcY + y);
-					gdImageSetPixel (dst, dstX + x, dstY + y, c);
+					if (c != src->transparent) {
+						gdImageSetPixel (dst, dstX + x, dstY + y, c);
+					}
 				}
 			}
 		} else {
@@ -2202,26 +2267,6 @@ void gdImageCopy (gdImagePtr dst, gdImagePtr src, int dstX, int dstY, int srcX,
 		return;
 	}
 
-	/* Destination is palette based */
-	if (src->trueColor) { /* But source is truecolor (Ouch!) */
-		toy = dstY;
-		for (y = srcY; (y < (srcY + h)); y++) {
-			tox = dstX;
-			for (x = srcX; x < (srcX + w); x++) {
-				int nc;
-				c = gdImageGetPixel (src, x, y);
-
-				/* Get best match possible. */
-				nc = gdImageColorResolveAlpha(dst, gdTrueColorGetRed(c), gdTrueColorGetGreen(c), gdTrueColorGetBlue(c), gdTrueColorGetAlpha(c));
-
-				gdImageSetPixel(dst, tox, toy, nc);
-				tox++;
-			}
-			toy++;
-		}
-		return;
-	}
-
 	/* Palette based to palette based */
 	for (i = 0; i < gdMaxColors; i++) {
 		colorMap[i] = (-1);
@@ -2573,42 +2618,33 @@ void gdImageCopyResampled (gdImagePtr dst, gdImagePtr src, int dstX, int dstY, i
 }
 
 void gdImagePolygon (gdImagePtr im, gdPointPtr p, int n, int c)
+{
+	if (n <= 0) {
+		return;
+	}
+
+
+	gdImageLine (im, p->x, p->y, p[n - 1].x, p[n - 1].y, c);
+	gdImageOpenPolygon (im, p, n, c);
+}
+
+void gdImageOpenPolygon (gdImagePtr im, gdPointPtr p, int n, int c)
 {
 	int i;
 	int lx, ly;
-	typedef void (*image_line)(gdImagePtr im, int x1, int y1, int x2, int y2, int color);
-	image_line draw_line;
 
 	if (n <= 0) {
 		return;
 	}
 
-	/* Let it be known that we are drawing a polygon so that the opacity
-	 * mask doesn't get cleared after each line.
-	 */
-	if (c == gdAntiAliased) {
-		im->AA_polygon = 1;
-	}
-
-	if ( im->antialias) {
-		draw_line = gdImageAALine;
-	} else {
-		draw_line = gdImageLine;
-	}
 	lx = p->x;
 	ly = p->y;
-	draw_line(im, lx, ly, p[n - 1].x, p[n - 1].y, c);
 	for (i = 1; i < n; i++) {
 		p++;
-		draw_line(im, lx, ly, p->x, p->y, c);
+		gdImageLine(im, lx, ly, p->x, p->y, c);
 		lx = p->x;
 		ly = p->y;
 	}
-
-	if (c == gdAntiAliased) {
-		im->AA_polygon = 0;
-		gdImageAABlend(im);
-	}
 }
 
 int gdCompareInt (const void *a, const void *b);
@@ -2669,6 +2705,19 @@ void gdImageFilledPolygon (gdImagePtr im, gdPointPtr p, int n, int c)
 			maxy = p[i].y;
 		}
 	}
+	/* necessary special case: horizontal line */
+	if (n > 1 && miny == maxy) {
+		x1 = x2 = p[0].x;
+		for (i = 1; (i < n); i++) {
+			if (p[i].x < x1) {
+				x1 = p[i].x;
+			} else if (p[i].x > x2) {
+				x2 = p[i].x;
+			}
+		}
+		gdImageLine(im, x1, miny, x2, miny, c);
+		return;
+	}
 	pmaxy = maxy;
 	/* 2.0.16: Optimization by Ilia Chipitsine -- don't waste time offscreen */
 	if (miny < 0) {
@@ -2738,6 +2787,9 @@ void gdImageSetStyle (gdImagePtr im, int *style, int noOfPixels)
 	if (im->style) {
 		gdFree(im->style);
 	}
+	if (overflow2(sizeof (int), noOfPixels)) {
+		return;
+	}
 	im->style = (int *) gdMalloc(sizeof(int) * noOfPixels);
 	memcpy(im->style, style, sizeof(int) * noOfPixels);
 	im->styleLength = noOfPixels;
@@ -2867,29 +2919,6 @@ int gdImageCompare (gdImagePtr im1, gdImagePtr im2)
 	return cmpStatus;
 }
 
-int
-gdAlphaBlendOld (int dst, int src)
-{
-	/* 2.0.12: TBB: alpha in the destination should be a
-	 * component of the result. Thanks to Frank Warmerdam for
-	 * pointing out the issue.
-	 */
-	return ((((gdTrueColorGetAlpha (src) *
-	     gdTrueColorGetAlpha (dst)) / gdAlphaMax) << 24) +
-	  ((((gdAlphaTransparent - gdTrueColorGetAlpha (src)) *
-	     gdTrueColorGetRed (src) / gdAlphaMax) +
-	    (gdTrueColorGetAlpha (src) *
-	     gdTrueColorGetRed (dst)) / gdAlphaMax) << 16) +
-	  ((((gdAlphaTransparent - gdTrueColorGetAlpha (src)) *
-	     gdTrueColorGetGreen (src) / gdAlphaMax) +
-	    (gdTrueColorGetAlpha (src) *
-	     gdTrueColorGetGreen (dst)) / gdAlphaMax) << 8) +
-	  (((gdAlphaTransparent - gdTrueColorGetAlpha (src)) *
-	    gdTrueColorGetBlue (src) / gdAlphaMax) +
-	   (gdTrueColorGetAlpha (src) *
-	    gdTrueColorGetBlue (dst)) / gdAlphaMax));
-}
-
 int gdAlphaBlend (int dst, int src) {
     int src_alpha = gdTrueColorGetAlpha(src);
     int dst_alpha, alpha, red, green, blue;
@@ -2940,19 +2969,12 @@ void gdImageAlphaBlending (gdImagePtr im, int alphaBlendingArg)
 	im->alphaBlendingFlag = alphaBlendingArg;
 }
 
-void gdImageAntialias (gdImagePtr im, int antialias)
-{
-	if (im->trueColor){
-		im->antialias = antialias;
-	}
-}
-
 void gdImageSaveAlpha (gdImagePtr im, int saveAlphaArg)
 {
 	im->saveAlphaFlag = saveAlphaArg;
 }
 
-static int gdLayerOverlay (int dst, int src)
+int gdLayerOverlay (int dst, int src)
 {
 	int a1, a2;
 	a1 = gdAlphaMax - gdTrueColorGetAlpha(dst);
@@ -2985,6 +3007,28 @@ static int gdAlphaOverlayColor (int src, int dst, int max )
 	}
 }
 
+int gdLayerMultiply (int dst, int src)
+{
+	int a1, a2, r1, r2, g1, g2, b1, b2;
+	a1 = gdAlphaMax - gdTrueColorGetAlpha(src);
+	a2 = gdAlphaMax - gdTrueColorGetAlpha(dst);
+
+	r1 = gdRedMax - (a1 * (gdRedMax - gdTrueColorGetRed(src))) / gdAlphaMax;
+	r2 = gdRedMax - (a2 * (gdRedMax - gdTrueColorGetRed(dst))) / gdAlphaMax;
+	g1 = gdGreenMax - (a1 * (gdGreenMax - gdTrueColorGetGreen(src))) / gdAlphaMax;
+	g2 = gdGreenMax - (a2 * (gdGreenMax - gdTrueColorGetGreen(dst))) / gdAlphaMax;
+	b1 = gdBlueMax - (a1 * (gdBlueMax - gdTrueColorGetBlue(src))) / gdAlphaMax;
+	b2 = gdBlueMax - (a2 * (gdBlueMax - gdTrueColorGetBlue(dst))) / gdAlphaMax ;
+
+	a1 = gdAlphaMax - a1;
+	a2 = gdAlphaMax - a2;
+	return ( ((a1*a2/gdAlphaMax) << 24) +
+			 ((r1*r2/gdRedMax) << 16) +
+			 ((g1*g2/gdGreenMax) << 8) +
+			 ((b1*b2/gdBlueMax))
+		);
+}
+
 void gdImageSetClip (gdImagePtr im, int x1, int y1, int x2, int y2)
 {
 	if (x1 < 0) {
@@ -3025,6 +3069,12 @@ void gdImageGetClip (gdImagePtr im, int *x1P, int *y1P, int *x2P, int *y2P)
 	*y2P = im->cy2;
 }
 
+void gdImageSetResolution(gdImagePtr im, const unsigned int res_x, const unsigned int res_y)
+{
+	if (res_x > 0) im->res_x = res_x;
+	if (res_y > 0) im->res_y = res_y;
+}
+
 /* convert a palette image to true color */
 int gdImagePaletteToTrueColor(gdImagePtr src)
 {
